// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_move.h"

#include "bhv_basic_tackle.h"
#include "bhv_receive_pass.h"
#include "bhv_mark_player.h"
#include "bhv_defense_mark.h"

#include "pass_course_table.h"

#include <rcsc/action/basic_actions.h>
#include "body_go_to_point_devel.h"
#if 0
#include <rcsc/action/body_intercept.h>
#else
#include "body_intercept_devel.h"
#endif
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include "body_press.h"
#include "body_get_ball.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/player/abstract_player_object.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#define DO_LOCAL_POSITIONING
#if defined(DO_LOCAL_POSITIONING)
# include "local_positioning.h"
#endif

#if 0
#define DO_MARK_PLAYER
#endif

#if 0
// doesn't work?
#define DO_DEFENSE_LINE_MARK
#endif

#if 1
#define DO_PRESS
#endif

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_BasicMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_BasicMove"
                        ,__FILE__, __LINE__ );

    // tackle

    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return true;
    }

    const rcsc::WorldModel & wm = agent->world();

    // check ball owner

#if 1
    int self_min = wm.interceptTable()->selfReachCycle();
#else
    int self_min = Body_GetBall::gettable_step( wm );
#endif
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    const int our_team_min = std::min( wm.interceptTable()->teammateReachCycle(),
                                       wm.interceptTable()->selfReachCycle() );
    const int their_team_min = wm.interceptTable()->opponentReachCycle();

    const int ball_reach_step = std::min( our_team_min, their_team_min );

    const rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( ball_reach_step );

    if ( ! wm.existKickableTeammate()
         && ( self_min <= 3
              || ( self_min < mate_min + 3
                   && self_min < opp_min + 4 )
              )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: receive pass"
                            ,__FILE__, __LINE__ );

        if ( Bhv_ReceivePass().execute( agent ) )
        {
            return true;
        }
    }


#if defined(DO_PRESS)
    //
    // press
    //
    if ( wm.interceptTable()->teammateReachCycle()
         > wm.interceptTable()->opponentReachCycle()
         && wm.self().pos().x > ball_pos.x
         && std::fabs( M_home_pos.y - wm.self().pos().x ) <= 5.0 )
    {
        if ( Body_Press().execute( agent ) )
        {
             agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

             return true;
        }
    }
#endif

    // go to home position

    const double dash_power = getDashPower( agent, M_home_pos );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

#if defined(DO_LOCAL_POSITIONING)
    puppets::LocalPositioning localPos;
    rcsc::Vector2D positioningTarget = M_home_pos;
    if ( (wm.self().inertiaPoint( 1 ) - M_home_pos).r() < 2.0 )
    {
        positioningTarget = localPos.getLocalAdjustmentPoint( agent, M_home_pos );
    }
#else
    const rcsc::Vector2D positioningTarget = M_home_pos;
#endif

#if defined(DO_MARK_PLAYER)
    if ( opp_min <= std::min( mate_min, self_min )
         && M_home_pos.x <= rcsc::ServerParam::i().ourPenaltyAreaLineX()
         && std::fabs( M_home_pos.y ) <= rcsc::ServerParam::i().penaltyAreaHalfWidth() )
    {
        rcsc::AbstractPlayerCont free_opponents = wm.getPlayerCont
            ( new rcsc::AndPlayerPredicate
              ( new rcsc::CoordinateAccuratePlayerPredicate( 10 ),
                new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                new rcsc::PointNearPlayerPredicate( M_home_pos, 10.0 ),
                new rcsc::NotPlayerPredicate
                ( new rcsc::ExistNearPlayerPlayerPredicate
                  ( wm,
                    new rcsc::TeammateOrSelfPlayerPredicate( wm ),
                    2.0 ) ) ) );

        // XXX: sort
        if ( ! free_opponents.empty() )
        {
            return Bhv_MarkPlayer( *(free_opponents.begin()) ).execute( agent );
        }
    }
#endif

#if defined(DO_DEFENSE_LINE_MARK)
    if ( M_home_pos.x >= -30.0
         && M_home_pos.x <= wm.defenseLineX() + 10.0 )
    {
        // find free opponent near my position
        rcsc::AbstractPlayerCont free_opponents = wm.getPlayerCont
            ( new rcsc::AndPlayerPredicate
              ( new rcsc::CoordinateAccuratePlayerPredicate( 20 ),
                new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                new rcsc::XCoordinateBackwardPlayerPredicate
                    ( wm.defenseLineX() + 20.0 ),
                new rcsc::PointNearPlayerPredicate( M_home_pos, 20.0 ),
                new rcsc::NotPlayerPredicate
                    ( new rcsc::ExistNearPlayerPlayerPredicate
                      ( wm,
                        new rcsc::TeammateOrSelfPlayerPredicate( wm ),
                        2.0 ) ) ) );

        // XXX: sort
        if ( ! free_opponents.empty() )
        {
            Bhv_DefenseMark( *(free_opponents.begin()) ).execute( agent );
            return true;
        }
    }
#endif

    //
    // avoid offside
    //
    if ( wm.self().pos().x > wm.offsideLineX() )
    {
        agent->debugClient().addMessage( "AvoidOffside" );

        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

        return Body_GoToPoint_Devel( wm.self().pos()
                                     + rcsc::Vector2D( -10.0, 0.0 ) )
                                   .execute( agent );
    }

    //
    // go forward
    //
    if ( wm.interceptTable()->teammateReachCycle()
         < wm.interceptTable()->opponentReachCycle()
         && ball_pos.x > 25.0
         && (positioningTarget - wm.self().pos()).r() > 10.0
         && wm.self().pos().x < wm.offsideLineX() - 5 )
    {
        agent->debugClient().addMessage( "goForward" );

        Body_GoToPoint_Devel( rcsc::Vector2D( positioningTarget.x,
                                              wm.self().pos().y ),
                              5.0, rcsc::ServerParam::i().maxPower() )
            .execute( agent );

        return true;
    }

    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );

    if ( ! Body_GoToPoint_Devel( positioningTarget, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );

        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

        return true;
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
double
Bhv_BasicMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    const int our_team_min = std::min( wm.interceptTable()->teammateReachCycle(),
                                       wm.interceptTable()->selfReachCycle() );
    const int their_team_min = wm.interceptTable()->opponentReachCycle();

    const int ball_reach_step = std::min( our_team_min, their_team_min );

    const rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( ball_reach_step );

#if 1
    if ( wm.self().stamina()
         > rcsc::ServerParam::i().staminaMax()
           - wm.self().playerType().staminaIncMax() * 5.0 )
    {
        // around max stamina, consider to some feature turns and cut off
        // by stamina max, use max power.
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else
#endif
    if ( wm.defenseLineX() > wm.self().pos().x
         && ball_pos.x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: correct DF line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
#if 1
    else if ( target_point.x < rcsc::ServerParam::i().ourPenaltyAreaLineX() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: positioning target too back. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( ball_pos.x < wm.defenseLineX() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: ball will be behind of defense line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( ball_pos.x < wm.defenseLineX() - 5.0
              && their_team_min <= our_team_min )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: ball is near defense line and exist kickable opponent. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( ball_pos.x > wm.offsideLineX() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: ball will be forward of offside line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( target_point.x < rcsc::ServerParam::i().ourPenaltyAreaLineX() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: positioning target too back. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
# if 1
    else if ( target_point.x
              > rcsc::ServerParam::i().theirPenaltyAreaLineX() - 10.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: positioning target too forward. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
# endif
#endif
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // prefered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable teammate. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: normal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }


    // if cut off by max stamina threshold will occur, use as much as possible
    if ( wm.self().stamina() + wm.self().playerType().staminaIncMax()
         - rcsc::ServerParam::i().staminaMax() > dash_power )
    {
        double new_dash_power = wm.self().stamina()
                                + wm.self().playerType().staminaIncMax()
                                - rcsc::ServerParam::i().staminaMax();

        dash_power = new_dash_power;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: updated power not to cut off from %f to %f"
                            ,__FILE__, __LINE__,
                            dash_power, new_dash_power);
    }

    return dash_power;
}
