// -*-c++-*-

/*!
  \file bhv_receive_pass.cpp
  \brief run to ball with checking information of players for next pass.
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_receive_pass.h"

#include "pass_course_table.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/soccer_action.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#if 0
#include <rcsc/action/body_intercept.h>
#else
#include "body_intercept_devel.h"
#endif
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/view_synch.h>
#include <rcsc/action/view_wide.h>

/*-------------------------------------------------------------------*/
/*!

*/
Bhv_ReceivePass::Bhv_ReceivePass()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_ReceivePass::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__ ": execute()" );

    const rcsc::WorldModel & wm = agent->world();


    //
    // execute intercept
    //
    rcsc::Body_Intercept().execute( agent );


    //
    // execute neck and view action
    //
    const PassCourseTable & table = PassCourseTable::instance( wm, true );
    const PassCourseTable::CourseSelection pass_candidate
          = table.get_first_selection( agent );

    int rest_cycles = wm.interceptTable()->selfReachCycle();
    if ( agent->effector().queuedNextBallKickable() )
    {
        rest_cycles = 1;
    }

    const rcsc::SeeState & see_state = agent->seeState();

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "rest cycles = %d", rest_cycles );
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "next see = %d", see_state.cyclesTillNextSee() );

    int see_cycles = see_state.cyclesTillNextSee() + 1;

    if ( rest_cycles == 4
         && wm.ball().posCount() <= 1 )
    {
        agent->setViewAction( new rcsc::View_Wide() );
    }
    else if ( rest_cycles < see_cycles )
    {
        agent->setViewAction( new rcsc::View_Synch() );
    }


    rcsc::NeckAction * neck_action = static_cast<rcsc::NeckAction *>( 0 );
    switch( pass_candidate.type )
    {
    case PassCourseTable::CourseSelection::Shoot:
    case PassCourseTable::CourseSelection::Push: /* NOTREACHED */
    case PassCourseTable::CourseSelection::Hold: /* NOTREACHED */
        neck_action = new rcsc::Neck_TurnToLowConfTeammate();
        break;

    case PassCourseTable::CourseSelection::FindPassReceiver:
    case PassCourseTable::CourseSelection::Pass:
        {
            const rcsc::AbstractPlayerObject * target;
            target = wm.teammate( pass_candidate.player_number );

            const rcsc::Vector2D target_pos = target->inertiaPoint( 1 );

            if ( target )
            {
                agent->debugClient().addMessage( "precheck find" );
                agent->debugClient().addLine
                    ( agent->effector().queuedNextMyPos(),
                      target_pos );
            }

            if ( rest_cycles == 1
                 || (see_cycles == 1
                     && wm.self().viewWidth() == rcsc::ViewWidth::WIDE) )
            {
                neck_action = new rcsc::Neck_TurnToPoint( target_pos );
            }
            else
            {
                neck_action = new rcsc::Neck_TurnToPlayerOrScan( target );
            }

            agent->debugClient().setTarget( target -> unum() );

            break;
        }
    }

    agent->setNeckAction( neck_action );

    return true;
}
