// -*-c++-*-

/*!
  \file body_advance_ball2008.cpp
  \brief kick the ball to be as forward as possible.
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_advance_ball2008.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>

#include <rcsc/action/body_kick_multi_step.h>

#include "simulator.h"

#include <boost/shared_ptr.hpp>
#include <utility>
#include <algorithm>

#include "team_switch.h"
#if defined(USE_SMART_KICK)
# include "contrib/body_smart_kick.h"
#endif


using rcsc::EPS;

class KickPlan
{
private:
    boost::shared_ptr< rcsc::BodyAction > M_action;
    rcsc::Vector2D M_kick_result_pos;
    rcsc::Vector2D M_kick_result_vel;
    int M_kick_step;

public:
    KickPlan( const KickPlan & k )
        : M_action( k.getAction() )
        , M_kick_result_pos( k.getKickResultPos() )
        , M_kick_result_vel( k.getKickResultVel() )
        , M_kick_step( k.getKickStep() )
      {
      }

    KickPlan( const boost::shared_ptr< rcsc::BodyAction > & action,
              const rcsc::Vector2D & kick_result_pos,
              const rcsc::Vector2D & kick_result_vel,
              int kick_step )
        : M_action( action )
        , M_kick_result_pos( kick_result_pos )
        , M_kick_result_vel( kick_result_vel )
        , M_kick_step( kick_step )
      {
      }

    bool isValid() const
      {
          return M_action;
      }

    boost::shared_ptr< rcsc::BodyAction > getAction() const
      {
          return M_action;
      }

    const rcsc::Vector2D & getKickResultPos() const
      {
          return M_kick_result_pos;
      }

    const rcsc::Vector2D & getKickResultVel() const
      {
          return M_kick_result_vel;
      }

    int getKickStep() const
      {
          return M_kick_step;
      }
};

typedef std::vector< KickPlan > KickPlanCont;

struct SimulatorEstimatedSituation
{
    KickPlan plan;

    long self_access_step;
    long teammate_access_step;
    long opponent_access_step;
    long unknown_access_step;
    int first_access_teammate;
    int first_access_opponent;

    SimulatorEstimatedSituation( const KickPlan & plan,
                                 long self_access_step,
                                 long teammate_access_step,
                                 long opponent_access_step,
                                 long unknown_access_step,
                                 int first_access_teammate,
                                 int first_access_opponent )
        : plan( plan )
        , self_access_step( self_access_step )
        , teammate_access_step( teammate_access_step )
        , opponent_access_step( opponent_access_step )
        , unknown_access_step( unknown_access_step )
        , first_access_teammate( first_access_teammate )
        , first_access_opponent( first_access_opponent )
      {
      }
};

static
boost::shared_ptr< KickPlan >
getBestChoice( std::pair< bool, double >
                 (*evaluator)( const SimulatorEstimatedSituation & s,
                               rcsc::PlayerAgent * agent ),
               KickPlanCont & candidates,
               const rcsc::AbstractPlayerCont & players,
               rcsc::PlayerAgent * agent );

static
std::pair< bool, double >
evaluateSituation( const SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent );

/*-------------------------------------------------------------------*/
/*!

*/
Body_AdvanceBall2008::Body_AdvanceBall2008()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_AdvanceBall2008::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();


    //
    // create candidates
    //
    KickPlanCont kick_candidates;

    for ( double theta = -80.0; theta <= +80.0 + EPS; theta += 5.0 )
    {
        const rcsc::Vector2D target( wm.self().pos()
                                     + rcsc::Vector2D::polar2vector
                                                       ( 100.0, theta ) );

        const rcsc::AngleDeg th( theta );

        double target_ball_speed = param.ballSpeedMax();
        const double cos_th = th.cos();

        if ( cos_th >= EPS )
        {
            const double x_rest = std::min( param.theirTeamGoalLineX() - 10.0,
                                            wm.self().pos().x )
                                  - wm.self().pos().x;

            const double max_dist = x_rest / cos_th;

            target_ball_speed = std::min( param.ballSpeedMax(),
                                          rcsc::calc_first_term_inf_geom_series
                                          ( max_dist, param.ballDecay() ) );
        }

        const rcsc::Vector2D result_pos = wm.self().pos();
        const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                          ( target_ball_speed, th );
        const int kick_step = 2;


        kick_candidates.push_back( KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
#if defined(USE_SMART_KICK)
                                     ( new rcsc::Body_SmartKick
                                       ( target,
                                         param.ballSpeedMax(),
                                         param.ballSpeedMax() * 0.96,
                                         3 ) ),
#else
                                     ( new rcsc::Body_KickMultiStep
                                       ( target,
                                         param.ballSpeedMax(),
                                         true ) ),
#endif
                                     result_pos,
                                     result_vel,
                                     kick_step ) );
    }


    //
    // get the best plan
    //
    boost::shared_ptr< KickPlan > best;
    best = getBestChoice( evaluateSituation,
                          kick_candidates,
                          wm.getPlayerCont
                          ( new rcsc::CoordinateAccuratePlayerPredicate
                            ( 10 ) ),
                          agent );

    //
    // execute
    //
    if ( best )
    {
        return best->getAction()->execute( agent );
    }
    else
    {
        return false;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
static
boost::shared_ptr< KickPlan >
getBestChoice( std::pair< bool, double >
                 (*evaluator)( const SimulatorEstimatedSituation & s,
                               rcsc::PlayerAgent * agent ),
               KickPlanCont & candidates,
               const rcsc::AbstractPlayerCont & players,
               rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    Simulator sim;

    KickPlanCont::iterator best_plan = candidates.end();
    double max_eval = 0.0;

    KickPlanCont::const_iterator c_end = candidates.end();
    for ( KickPlanCont::iterator it = candidates.begin();
          it != c_end;
          ++it )
    {
        const int kick_step = (*it).getKickStep();

        const int self_penalty_step = kick_step;
        const int teammate_penalty_step = kick_step;
        const int opponent_penalty_step = kick_step;

        long self_access_step;
        long teammate_access_step;
        long opponent_access_step;
        long unknown_access_step;
        int first_access_teammate;
        int first_access_opponent;

        sim.simulate( wm,
                      players,
                      (*it).getKickResultPos(),
                      (*it).getKickResultVel(),
                      kick_step,
                      self_penalty_step,
                      teammate_penalty_step,
                      opponent_penalty_step,
                      &self_access_step,
                      &teammate_access_step,
                      &opponent_access_step,
                      &unknown_access_step,
                      &first_access_teammate,
                      &first_access_opponent );

        SimulatorEstimatedSituation sit( (*it),
                                         self_access_step,
                                         teammate_access_step,
                                         opponent_access_step,
                                         unknown_access_step,
                                         first_access_teammate,
                                         first_access_opponent );

        std::pair< bool, double > eval = evaluateSituation( sit, agent );

        if ( eval.first )
        {
            if ( max_eval < eval.second 
                 || best_plan == candidates.end() )
            {
                max_eval = eval.second ;
                best_plan = it;
            }
        }
    }

    if ( best_plan == candidates.end() )
    {
        return boost::shared_ptr< KickPlan >
               ( static_cast<KickPlan *>(0) );
    }
    else
    {
        return boost::shared_ptr< KickPlan >
               ( new KickPlan( *best_plan ) );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
static
std::pair< bool, double >
evaluateSituation( const SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent )
{
    const rcsc::ServerParam & param = rcsc::ServerParam::i();
    const rcsc::WorldModel & wm = agent->world();

    long our_team_access_step = std::min( s.teammate_access_step,
                                          s.self_access_step );
    long opp_team_access_step = std::min( s.opponent_access_step,
                                          s.unknown_access_step );

    rcsc::Vector2D result_point = rcsc::inertia_n_step_point
                                        ( s.plan.getKickResultPos(),
                                          s.plan.getKickResultVel(),
                                          opp_team_access_step,
                                          param.ballDecay() );

    if ( result_point.x < - param.pitchHalfLength() + 2.0
         || result_point.x > param.pitchHalfLength() - 2.0
         || std::fabs( result_point.y ) > param.pitchHalfWidth() - 2.0 )
    {
        return std::make_pair( false, 0.0 );
    }


    if ( our_team_access_step <= opp_team_access_step )
    {
        agent->debugClient().addCircle( result_point, 0.5 );
    }
    else
    {
        const double d = 0.25;

        agent->debugClient().addLine
            ( result_point + rcsc::Vector2D( -d, -d ),
              result_point + rcsc::Vector2D( +d, +d ) );

        agent->debugClient().addLine
            ( result_point + rcsc::Vector2D( -d, +d ),
              result_point + rcsc::Vector2D( +d, -d ) );
    }


    //
    // set basic value
    //
    double value = result_point.x;


    //
    // add bonus if we get the ball
    //
    if ( our_team_access_step <= opp_team_access_step )
    {
        value += (opp_team_access_step - our_team_access_step + 1) * 100.0;
    }


    //
    // add bonus distance from result_point to first access oppenent
    //
    if ( s.first_access_opponent != rcsc::Unum_Unknown )
    {
        value += (result_point
                  - wm.opponent( s.first_access_opponent )->pos()).r();
    }

    return std::make_pair( true, value );
}
