// -*-c++-*-

/*!
  \file body_emergency_kick.cpp
  \brief kick the ball to keep ball by our team
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_emergency_kick.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <rcsc/action/body_kick_one_step.h>

#include "simulator.h"

#include <boost/shared_ptr.hpp>
#include <utility>
#include <algorithm>

static const double EPS = 1.0e-10;

class KickPlan
{
private:
    boost::shared_ptr< rcsc::BodyAction > M_action;
    rcsc::Vector2D M_kick_result_pos;
    rcsc::Vector2D M_kick_result_vel;
    int M_kick_step;

public:
    KickPlan( const KickPlan & k )
        : M_action( k.getAction() )
        , M_kick_result_pos( k.getKickResultPos() )
        , M_kick_result_vel( k.getKickResultVel() )
        , M_kick_step( k.getKickStep() )
      {
      }

    KickPlan( const boost::shared_ptr< rcsc::BodyAction > & action,
                   const rcsc::Vector2D & kick_result_pos,
                   const rcsc::Vector2D & kick_result_vel,
                   int kick_step )
        : M_action( action )
        , M_kick_result_pos( kick_result_pos )
        , M_kick_result_vel( kick_result_vel )
        , M_kick_step( kick_step )
      {
      }

    bool isValid() const
      {
          return M_action;
      }

    boost::shared_ptr< rcsc::BodyAction > getAction() const
      {
          return M_action;
      }

    const rcsc::Vector2D & getKickResultPos() const
      {
          return M_kick_result_pos;
      }

    const rcsc::Vector2D & getKickResultVel() const
      {
          return M_kick_result_vel;
      }

    int getKickStep() const
      {
          return M_kick_step;
      }
};

typedef std::vector< KickPlan > KickPlanCont;

struct SimulatorEstimatedSituation
{
    KickPlan plan;

    long self_access_step;
    long teammate_access_step;
    long opponent_access_step;
    long unknown_access_step;
    int first_access_teammate;
    int first_access_opponent;

    rcsc::Vector2D receive_point;

    SimulatorEstimatedSituation( const KickPlan & plan,
                                 long self_access_step,
                                 long teammate_access_step,
                                 long opponent_access_step,
                                 long unknown_access_step,
                                 int first_access_teammate,
                                 int first_access_opponent,
                                 const rcsc::Vector2D & receive_point )
        : plan( plan )
        , self_access_step( self_access_step )
        , teammate_access_step( teammate_access_step )
        , opponent_access_step( opponent_access_step )
        , unknown_access_step( unknown_access_step )
        , first_access_teammate( first_access_teammate )
        , first_access_opponent( first_access_opponent )
        , receive_point( receive_point )
      {
      }
};

static
boost::shared_ptr< SimulatorEstimatedSituation >
getBestChoice( std::pair< bool, double >
                 (*evaluator)( const SimulatorEstimatedSituation & s,
                               rcsc::PlayerAgent * agent ),
               KickPlanCont & candidates,
               const rcsc::AbstractPlayerCont & players,
               rcsc::PlayerAgent * agent );

static
std::pair< bool, double >
evaluateSituation( const SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent );

/*-------------------------------------------------------------------*/
/*!

*/
Body_EmergencyKick::Body_EmergencyKick()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_EmergencyKick::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const double ball_speed_max = rcsc::ServerParam::i().ballSpeedMax();

    //
    // create candidates
    //
    KickPlanCont kick_candidates;

    const rcsc::Vector2D ball_pos = wm.ball().pos();
    const rcsc::Vector2D ball_vel = wm.ball().vel();
    double kick_rate = wm.self().kickRate();


#if 1
    rcsc::PlayerCont::const_iterator t_end = wm.teammates().end();
    for ( rcsc::PlayerCont::const_iterator it = wm.teammates().begin();
          it != t_end;
          ++it )
    {
        const rcsc::Vector2D target( (*it).pos() + (*it).vel() );
        const rcsc::Vector2D & result_pos = ball_pos;
        const rcsc::Vector2D result_vel
            = rcsc::Body_KickOneStep::get_max_possible_vel
              ( (target - ball_pos).th(),
                kick_rate,
                ball_vel );
        const int kick_step = 1;

        kick_candidates.push_back( KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
                                     ( new rcsc::Body_KickOneStep
                                       ( target,
                                         result_vel.r() ) ),
                                     result_pos,
                                     result_vel,
                                     kick_step ) );
    }
#endif

    //
    // get the best plan
    //
    boost::shared_ptr< SimulatorEstimatedSituation > best;
    best = getBestChoice( evaluateSituation,
                          kick_candidates ,
                          wm.getPlayerCont
                          ( new rcsc::CoordinateAccuratePlayerPredicate
                            ( 10 ) ),
                          agent );

    //
    // execute
    //
    if ( best )
    {
        if ( best->plan.getAction()->execute( agent ) )
        {
            if ( best->first_access_teammate != wm.self().unum() )
            {
                rcsc::Vector2D target_buf = best->plan.getKickResultVel();
                target_buf.setLength( 1.0 );

                agent->addSayMessage
                       ( new rcsc::PassMessage
                         ( best->first_access_teammate,
                           best->receive_point + target_buf,
                           agent->effector().queuedNextBallPos(),
                           agent->effector().queuedNextBallVel() ) );
            }

            return true;
        }
    }


#if 1
    for ( double theta = -180.0; theta <= +180.0 + EPS; theta += 15.0 )
    {
        const rcsc::Vector2D target( ball_pos
                                     + rcsc::Vector2D::polar2vector
                                                       ( 10.0, theta ) );

        const rcsc::Vector2D result_pos = ball_pos;

        const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                          ( ball_speed_max,
                                            (target - ball_pos).th() );
        const int kick_step = 1;

#if 0
        kick_candidates.push_back( KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
                                     ( new rcsc::Body_KickOneStep
                                       ( target,
                                         param.ballSpeedMax() ) ),
                                     result_pos,
                                     result_vel,
                                     kick_step ) );
#endif

        kick_candidates.push_back( KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
                                     ( new rcsc::Body_KickOneStep
                                       ( target,
                                         ball_speed_max ) ),
                                     result_pos,
                                     result_vel * 0.75,
                                     kick_step ) );

        kick_candidates.push_back( KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
                                     ( new rcsc::Body_KickOneStep
                                       ( target,
                                         ball_speed_max ) ),
                                     result_pos,
                                     result_vel * 0.5,
                                     kick_step ) );
    }
#endif


    //
    // get the best plan
    //
    best = getBestChoice( evaluateSituation,
                          kick_candidates ,
                          wm.getPlayerCont
                          ( new rcsc::CoordinateAccuratePlayerPredicate
                            ( 10 ) ),
                          agent );

    //
    // execute
    //
    if ( best )
    {
        if ( best->plan.getAction()->execute( agent ) )
        {
            if ( best->first_access_teammate != wm.self().unum() )
            {
                rcsc::Vector2D target_buf = best->plan.getKickResultVel();
                target_buf.setLength( 1.0 );

                agent->addSayMessage
                       ( new rcsc::PassMessage
                         ( best->first_access_teammate,
                           best->receive_point + target_buf,
                           agent->effector().queuedNextBallPos(),
                           agent->effector().queuedNextBallVel() ) );
            }

            return true;
        }
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
static
boost::shared_ptr< SimulatorEstimatedSituation >
getBestChoice( std::pair< bool, double >
                 (*evaluator)( const SimulatorEstimatedSituation & s,
                               rcsc::PlayerAgent * agent ),
               KickPlanCont & candidates,
               const rcsc::AbstractPlayerCont & players,
               rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    Simulator sim;

    KickPlanCont::iterator best_plan = candidates.end();
    double max_eval = 0.0;
    boost::shared_ptr< SimulatorEstimatedSituation > best_situation;

    KickPlanCont::const_iterator c_end = candidates.end();
    for ( KickPlanCont::iterator it = candidates.begin();
          it != c_end;
          ++it )
    {
        const int kick_step = (*it).getKickStep();

        const int self_penalty_step = kick_step + 10;
        const int teammate_penalty_step = kick_step + 3;
        const int opponent_penalty_step = 1;

        long self_access_step;
        long teammate_access_step;
        long opponent_access_step;
        long unknown_access_step;
        int first_access_teammate;
        int first_access_opponent;

        sim.simulate( wm,
                      players,
                      (*it).getKickResultPos(),
                      (*it).getKickResultVel(),
                      kick_step,
                      self_penalty_step,
                      teammate_penalty_step,
                      opponent_penalty_step,
                      &self_access_step,
                      &teammate_access_step,
                      &opponent_access_step,
                      &unknown_access_step,
                      &first_access_teammate,
                      &first_access_opponent );

        long our_team_access_step = std::min( teammate_access_step,
                                              self_access_step );

        rcsc::Vector2D result_point = rcsc::inertia_n_step_point
                                        ( (*it).getKickResultPos(),
                                          (*it).getKickResultVel(),
                                          our_team_access_step,
                                          param.ballDecay() );
                
        boost::shared_ptr< SimulatorEstimatedSituation >
            sit( new SimulatorEstimatedSituation( (*it),
                                                  self_access_step,
                                                  teammate_access_step,
                                                  opponent_access_step,
                                                  unknown_access_step,
                                                  first_access_teammate,
                                                  first_access_opponent,
                                                  result_point ) );

        std::pair< bool, double > eval = evaluateSituation( *sit, agent );

        if ( eval.first )
        {
            if ( max_eval < eval.second 
                 || best_plan == candidates.end() )
            {
                max_eval = eval.second ;
                best_plan = it;
                best_situation = sit;
            }
        }
    }

    if ( best_plan == candidates.end() )
    {
        return boost::shared_ptr< SimulatorEstimatedSituation >
               ( static_cast< SimulatorEstimatedSituation * >( 0 ) );
    }
    else
    {
        return best_situation;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
static
std::pair< bool, double >
evaluateSituation( const SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    long our_team_access_step = std::min( s.teammate_access_step,
                                          s.self_access_step );
    long opp_team_access_step = std::min( s.opponent_access_step,
                                          s.unknown_access_step );

    rcsc::Vector2D result_point = rcsc::inertia_n_step_point
                                        ( s.plan.getKickResultPos(),
                                          s.plan.getKickResultVel(),
                                          our_team_access_step,
                                          param.ballDecay() );

    if ( result_point.x < - param.pitchHalfLength() + 2.0
         || result_point.x > param.pitchHalfLength() - 2.0
         || std::fabs( result_point.y ) > param.pitchHalfWidth() - 2.0 )
    {
        return std::make_pair( false, 0.0 );
    }


    if ( our_team_access_step <= opp_team_access_step )
    {
        agent->debugClient().addCircle( result_point, 0.5 );
    }
    else
    {
        const double d = 0.25;

        agent->debugClient().addLine
               ( result_point + rcsc::Vector2D( -d , -d ),
                 result_point + rcsc::Vector2D( +d , +d ) );

        agent->debugClient().addLine
               ( result_point + rcsc::Vector2D( -d , +d ),
                 result_point + rcsc::Vector2D( +d , -d ) );
    }


    //
    // set basic value
    //
    double value = result_point.x;


    //
    // add bonus if we get the ball
    //
    if ( our_team_access_step <= opp_team_access_step )
    {
        value += (opp_team_access_step - our_team_access_step + 1) * 10.0;
    }


    //
    // add bonus distance from result_point to first access oppenent
    //
    if ( s.first_access_opponent != rcsc::Unum_Unknown )
    {
        value += std::min
                 ( (result_point
                    - wm.opponent( s.first_access_opponent )->pos()).r(),
                   5.0 );
    }

    return std::make_pair( true, value );
}
