// -*-c++-*-

/*!
  \file body_get_ball.cpp
  \brief run to ball
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_get_ball.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>
#include "rcsc-ext/player/player_type_ext.h"

#include <algorithm>
#include <cmath>

Body_GetBall::CacheInfo Body_GetBall::S_cache;

/*-------------------------------------------------------------------*/
/*!

*/
Body_GetBall::Body_GetBall( double max_dist,
                            bool collide,
                            bool use_back_dash )
    : M_max_dist( max_dist )
    , M_collide( collide )
    , M_use_back_dash( use_back_dash )
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_GetBall::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    // if max_dist is default margin, overwrite it
    if ( M_max_dist < 0.0 )
    {
        M_max_dist = wm.self().kickableArea();
    }

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        "self pos [% -3.4f % -3.4f ](% -3.4f)\n",
                        wm.self().pos().x,
                        wm.self().pos().y,
                        wm.self().body().degree() );

    if ( wm.ball().distFromSelf() <= M_max_dist )
    {
        return( false );
    }

    S_cache.calculate( wm,
                       M_max_dist,
                       M_collide,
                       M_use_back_dash );

    switch( S_cache.M_command->type() )
    {
    case rcsc::PlayerCommand::DASH:
        agent->doDash( S_cache.M_command->dashPower() );
        break;

    case rcsc::PlayerCommand::TURN:
        agent->doTurn( S_cache.M_command->turnMoment() );
        break;

    default:
        /* NOTREACHED */
        agent->doTurn( 30.0 );
        break;
    }

    return true;
}

//
// gettable_step
//
long
Body_GetBall::gettable_step( const rcsc::WorldModel & wm,
                             double max_dist,
                             bool collide,
                             bool use_back_dash )
{
    Body_GetBall::S_cache.calculate( wm,
                                     max_dist,
                                     collide,
                                     use_back_dash );

    return Body_GetBall::S_cache.M_gettable_step;
}


//
// gettable_point
//
rcsc::Vector2D
Body_GetBall::gettable_point( const rcsc::WorldModel & wm,
                              double max_dist,
                              bool collide,
                              bool use_back_dash )
{
    Body_GetBall::S_cache.calculate( wm,
                                     max_dist,
                                     collide,
                                     use_back_dash );

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        "gettable_point = [%f,%f]",
                        Body_GetBall::S_cache.M_gettable_point.x,
                        Body_GetBall::S_cache.M_gettable_point.y );

    return( Body_GetBall::S_cache.M_gettable_point );
}


void
Body_GetBall::CacheInfo::calculate( const rcsc::WorldModel & wm,
                                    double max_dist,
                                    bool collide,
                                    bool use_back_dash )
{
    // if max_dist is default margin, overwrite it
    if ( max_dist < 0.0 )
    {
        max_dist = wm.self().kickableArea();
    }


    if ( wm.time() == M_time_stamp
         && std::fabs( max_dist - M_max_dist ) <= rcsc::EPS
         && collide == M_collide
         && use_back_dash == M_use_back_dash )
    {
        // cache is already generated, no need to calculate.
        return;
    }

#if 1
#define GET_BALL_FLAGMENT_TEST_DEBUG
#endif

#ifdef GET_BALL_FLAGMENT_TEST_DEBUG
    rcsc::dlog.addText( rcsc::Logger::ACTION, "Body_GetBall(cache)" );
#endif

    M_time_stamp = wm.time();
    M_max_dist = max_dist;
    M_collide = collide;
    M_use_back_dash = use_back_dash;


    for ( long t = 0; ; ++ t )
    {
        boost::shared_ptr<rcsc::PlayerBodyCommand> com;

        if ( spend_ball_time_for_ball_gettable
             ( wm, t, max_dist, collide, use_back_dash, &com )
             <= t )
        {
            M_command = com;
            M_gettable_step = t;
            M_gettable_point = wm.ball().inertiaPoint( t );

            return;
        }
    }
}


long
Body_GetBall::spend_ball_time_for_ball_gettable
                    ( const rcsc::WorldModel & wm,
                      long t,
                      double max_dist,
                      bool collide,
                      bool use_back_dash,
                      boost::shared_ptr<rcsc::PlayerBodyCommand> * command )
{
    // if max_dist is default margin, overwrite it
    if ( max_dist < 0.0 )
    {
        max_dist = wm.self().kickableArea();
    }

#ifdef GET_BALL_FLAGMENT_TEST_DEBUG
    rcsc::dlog.addText( rcsc::Logger::ACTION, "Body_GetBall(cache)" );
#endif

    const long time_buffer = 0;

    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    const double collision_dist = param.ballSize()
                                  + wm.self().playerType().playerSize();


#ifdef GET_BALL_FLAGMENT_TEST_DEBUG
    rcsc::dlog.addText( rcsc::Logger::ACTION, "t = %ld", t );
#endif

    rcsc::Vector2D ball_coord;
    rcsc::Vector2D self_coord;

    ball_coord = wm.ball().inertiaPoint( t );
    self_coord = wm.self().playerType().inertiaPoint
                                ( wm.self().pos(),
                                  wm.self().vel(),
                                  t );
    rcsc::Vector2D rel = (ball_coord - self_coord)
                         .rotatedVector( - wm.self().body() );


    rcsc::AngleDeg diff_angle = rel.th().abs();
    rcsc::AngleDeg cone_angle = std::fabs( rcsc::AngleDeg::rad2deg
                                           ( std::atan2( max_dist,
                                                         rel.r() ) ) );

    rcsc::AngleDeg min_cone_angle = 2.5 + std::max( rel.r(), 25.0 ) * 0.1;

    if ( cone_angle.degree() < min_cone_angle.degree() )
    {
        cone_angle = min_cone_angle;
    }


    bool back_dash_flag = false;
    bool back_turn_flag = false;
    long turn_step = 0;

    // distance have to run at least for step calculation
    // minus means no dash is needed.
    double dist_for_step = 0.0;

    // distance to run for calculating action
    double dist_for_action = 0.0;


    //
    // ball is forward or little side, run straight forward
    //
    if ( diff_angle.degree() < cone_angle.degree()
         || (rel.absY() < max_dist - 0.2 - rel.r() * min_cone_angle.sin()
             && rel.x >= 0.0) )
    {
#ifdef GET_BALL_FLAGMENT_TEST_DEBUG
        rcsc::dlog.addText( rcsc::Logger::ACTION, "straight forward" );
#endif

        turn_step = 0;

        dist_for_step = rel.x + 0.2;

        double sq = std::pow( max_dist, 2 ) - std::pow( rel.y, 2 );

        if ( sq > 0.0 )
        {
            dist_for_step -= std::sqrt( sq );
        }

        back_dash_flag = false;


        if ( collide )
        {
            dist_for_action = rel.x;
        }
        else
        {
            dist_for_action = rel.x;

            double sq = std::pow( collision_dist + 0.2, 2 )
                        - std::pow( rel.y, 2 );
            if ( sq > 0.0 )
            {
                dist_for_action -= std::sqrt( sq );
            }
        }
    }
    //
    // ball is backward or little side, run straight backward
    //
    else if ( use_back_dash
              && wm.gameMode().type() == rcsc::GameMode::PlayOn
              && rel.r() <= 10.0
              && wm.self().stamina() - 1.0
              >= std::max
              ( rcsc::ServerParam::i().staminaMax() / 2.0,
                rcsc::ServerParam::i().recoverDecThrValue() + 500.0 )
              && ((diff_angle + 180).abs() < cone_angle.degree()
                  || rel.absY()
                  < max_dist - 0.2 - rel.r() * min_cone_angle.sin()) )
    {
#ifdef GET_BALL_FLAGMENT_TEST_DEBUG
        rcsc::dlog.addText( rcsc::Logger::ACTION, "straight backward" );
#endif

        turn_step = 0;

        dist_for_step = std::fabs( rel.absX() ) + 0.2;

        double sq = std::pow( max_dist, 2 )
            - std::pow( rel.y, 2 );
        if ( sq > 0.0 )
        {
            dist_for_step -= std::sqrt( sq );
        }

        back_dash_flag = true;


        if ( collide )
        {
            dist_for_action = rel.absX();
        }
        else
        {
            dist_for_action = rel.absX();

            double sq = std::pow( collision_dist + 0.2, 2 ) - std::pow( rel.y, 2 );
            if ( sq > 0.0 )
            {
                dist_for_action -= std::sqrt( sq );
            }
        }
    }
    //
    // turn needed
    //
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION, "turn needed" );

        long normal_turn_step
            = rcsc_ext::PlayerType_cyclesToTurnBody
            ( wm.self().playerType(),
              diff_angle.degree(), wm.self().vel().r() );

        long back_turn_step
            = rcsc_ext::PlayerType_cyclesToTurnBody
            ( wm.self().playerType(),
              (rel.th() + 180).abs(),
              wm.self().vel().r() );

        if ( back_turn_step < normal_turn_step
             && use_back_dash
             && wm.gameMode().type() == rcsc::GameMode::PlayOn
             && (wm.ball().inertiaPoint( t + back_turn_step )
                 - wm.self().playerType().inertiaPoint
                 ( wm.self().pos(),
                   wm.self().vel(),
                   t + back_turn_step )).r() <= 10.0
             && wm.self().stamina() - 1.0
             >= std::max
             ( rcsc::ServerParam::i().staminaMax() / 2.0,
               rcsc::ServerParam::i().recoverDecThrValue() + 500.0 )
             && t <= 3 )
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION, "back turn" );

            back_turn_flag = true;
            turn_step = back_turn_step;
        }
        else
        {
            back_turn_flag = false;
            turn_step = normal_turn_step;
        }


        dist_for_step = rel.r() - max_dist + 0.2;

        // unused
        if ( collide )
        {
            dist_for_action = rel.r();
        }
        else
        {
            dist_for_action = rel.r() - collision_dist;
        }
    }


    // XXX: override below using stop and turn
    if ( turn_step >= 3 )
    {
        turn_step = 2;
    }


    long spend_time = 0;

    spend_time += turn_step;

    if ( dist_for_step > 0.0 )
    {
        spend_time += wm.self().playerType()
            .cyclesToReachDistance( dist_for_step );
    }


    rcsc::dlog.addText( rcsc::Logger::ACTION, "rel = [%f,%f]", rel.x, rel.y );
    rcsc::dlog.addText( rcsc::Logger::ACTION, "turn_step = %ld", turn_step );
    rcsc::dlog.addText( rcsc::Logger::ACTION, "dist_for_step = %ld", dist_for_step );
    rcsc::dlog.addText( rcsc::Logger::ACTION, "dist_for_action = %f", dist_for_action );
    rcsc::dlog.addText( rcsc::Logger::ACTION, "spend_time = %ld", spend_time );

    boost::shared_ptr<rcsc::PlayerBodyCommand> com;

    if ( turn_step >= 1 )
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION, "turn needed" );

        if ( turn_step >= 3 )
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION, "stop" );

            double d_pow;
            d_pow = rcsc_ext::PlayerType_getStopPower
                ( wm.self().playerType(),
                  wm.self().vel().r() );

            com = boost::shared_ptr<rcsc::PlayerBodyCommand>
                  ( new rcsc::PlayerDashCommand
                    ( wm.self().getSafetyDashPower( d_pow ) ) );
        }
        else
        {
            rcsc::AngleDeg turn_angle;

            if ( back_turn_flag )
            {
                turn_angle = rel.th() + 180.0;
            }
            else
            {
                turn_angle = rel.th();
            }

#if 0
            com = boost::shared_ptr<rcsc::PlayerBodyCommand>
                ( new rcsc::PlayerTurnCommand
                  ( rcsc_ext::PlayerType_momentForTurn
                    ( wm.self().playerType(),
                      turn_angle.degree(),
                      wm.self().vel().r() ) ) );
#else
            com = boost::shared_ptr<rcsc::PlayerBodyCommand>
                  ( new rcsc::PlayerTurnCommand( turn_angle.degree() ) );
#endif
        }
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION, "turn not needed" );

        double dash_power
            = rcsc_ext::PlayerType_getDashPowerToReachNextCycle
              ( wm.self().playerType(),
                dist_for_action * (back_dash_flag ? -1 : 1) );

        com = boost::shared_ptr<rcsc::PlayerBodyCommand>
            ( new rcsc::PlayerDashCommand
              ( wm.self().getSafetyDashPower( dash_power ) ) );
    }


    if ( command )
    {
        *(command) = com;
    }

    return spend_time + std::min( time_buffer, spend_time );
}
