// -*-c++-*-

/*!
  \file body_go_to_forward_free_space.cpp
  \brief go to free point near opponent's offside line
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_go_to_forward_free_space.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/common/logger.h>
#include "rcsc-ext/player/world_model_ext.h"
#include "rcsc-ext/player/player_evaluator.h"
#include <set>

#include "body_go_to_point_devel.h"


/*-------------------------------------------------------------------*/
/*!

*/
Body_GoToForwardFreeSpace::Body_GoToForwardFreeSpace( double y_min,
                                                      double y_max,
                                                      double dash_power )
    : M_y_min( y_min )
    , M_y_max( y_max )
    , M_dash_power( dash_power )
{
    if ( y_min > y_max )
    {
        M_y_max = y_min;
        M_y_min = y_max;
    }

    if ( M_y_min < - rcsc::ServerParam::i().pitchHalfWidth() )
    {
        M_y_min = - rcsc::ServerParam::i().pitchHalfWidth();
    }

    if ( M_y_max > + rcsc::ServerParam::i().pitchHalfWidth() )
    {
        M_y_max = + rcsc::ServerParam::i().pitchHalfWidth();
    }

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": y_min = %f, y_max = %f",
                        M_y_min, M_y_max );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_GoToForwardFreeSpace::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    double pos_x = this->getPositionX( wm );

    // TODO: check out of field
    rcsc::AbstractPlayerCont defenders = wm.getPlayerCont
        ( new rcsc::AndPlayerPredicate
          ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
            new rcsc::XCoordinateForwardPlayerPredicate
                ( wm.offsideLineX() - 3.0 ),
            new rcsc::XCoordinateBackwardPlayerPredicate
                ( wm.offsideLineX() + 3.0 ),
            new rcsc::YCoordinatePlusPlayerPredicate( M_y_min ),
            new rcsc::YCoordinateMinusPlayerPredicate( M_y_max ) ) );

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": number of defenders = %d",
                        defenders.size() );

    bool found = false;
    double max_free_width = -1.0;
    double best_y = (M_y_min + M_y_max) / 2.0;

    std::set< double > defenders_y;

    const rcsc::AbstractPlayerCont::const_iterator d_end = defenders.end();
    for ( rcsc::AbstractPlayerCont::const_iterator it = defenders.begin();
          it != d_end;
          ++ it )
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": defenders y = %f",
                            (**it).pos().y );

        defenders_y.insert( (**it).pos().y );
    }

    const std::set< double >::const_iterator y_end = defenders_y.end();
    for ( std::set< double >::const_iterator it = defenders_y.begin();
          it != y_end;
          ++ it )
    {
        std::set< double >::const_iterator next = it;
        ++ next;

        if ( next == y_end )
        {
            break;
        }

        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": p1 y = %f, p2 y = %f",
                            *it, *next );

        double free_width = ((*next) - (*it)) / 2.0;

        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": free_width = %f",
                            free_width );

        if ( free_width > max_free_width )
        {
            found = true;
            max_free_width = free_width;
            best_y = ((*it) + (*next)) / 2.0;

            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                __FILE__ ": best_y = %f",
                                best_y );
        }
    }

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": best_y candidate between defenders = %f",
                        best_y );

    {
        rcsc::AbstractPlayerCont all_defenders = wm.getPlayerCont
            ( new rcsc::AndPlayerPredicate
              (
#if 0
                new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
#else
                new rcsc::NotPlayerPredicate
                ( new rcsc::SelfPlayerPredicate( wm ) ),
#endif
                new rcsc::XCoordinateForwardPlayerPredicate
                    ( pos_x - 3.0 ),
                new rcsc::XCoordinateBackwardPlayerPredicate
                    ( wm.offsideLineX() + 3.0 ) ) );

        double width_y_min = AbstractPlayerCont_getMinimumEvaluation
                             ( all_defenders,
                               new rcsc_ext::AbsYDiffPlayerEvaluator
                               ( rcsc::Vector2D( pos_x,
                                                 M_y_min ) ) );

        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": width_y_min = %f",
                            width_y_min );

        if ( width_y_min > max_free_width )
        {
            max_free_width = width_y_min;
            best_y = M_y_min;
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                __FILE__ ": updated best_y = %f",
                                best_y );
        }

        double width_y_max = AbstractPlayerCont_getMinimumEvaluation
                             ( all_defenders,
                               new rcsc_ext::AbsYDiffPlayerEvaluator
                               ( rcsc::Vector2D( pos_x,
                                                 M_y_max ) ) );

        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": width_y_max = %f",
                            width_y_max );

        if ( width_y_max > max_free_width )
        {
            max_free_width = width_y_max;
            best_y = M_y_max;
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                __FILE__ ": updated best_y = %f",
                                best_y );
        }
    }

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": best_y = %f",
                        best_y );

    rcsc::Vector2D target_point( pos_x, best_y );
    agent->debugClient().setTarget( target_point );

    return Body_GoToPoint_Devel( target_point, 10.0 , M_dash_power )
           .execute( agent )
        || Body_GoToPoint_Devel( target_point, 5.0 , M_dash_power )
           .execute( agent )
        || Body_GoToPoint_Devel( target_point, 2.0 , M_dash_power )
           .execute( agent )
        || Body_GoToPoint_Devel( target_point, 1.0 , M_dash_power )
           .execute( agent );
}

double
Body_GoToForwardFreeSpace::getPositionX( const rcsc::WorldModel & wm ) const
{
    const double offside_buffer = 1.0;

    double pos_x = wm.offsideLineX() - offside_buffer;

    if ( pos_x > wm.defenseLineX() + 60.0 )
    {
        pos_x = wm.defenseLineX() + 60.0;
    }

    return pos_x;
}
