// -*-c++-*-

/*!
  \file kick_simulator.h
  \brief kick simulator class Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_PLAYER_KICK_SIMULATOR_H
#define RCSC_PLAYER_KICK_SIMULATOR_H

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/angle_deg.h>

#include <vector>
#include <algorithm>

namespace rcsc {

class GameTime;
class PlayerType;
class WorldModel;

/*-------------------------------------------------------------------*/
/*!
  \class KickSimulator
  \brief imulator to generate smart kick.
*/
class KickSimulator {
public:

    static const double NEAR_SIDE_RATE;
    static const double MID_RATE;
    static const double FAR_SIDE_RATE;

    enum {
        MAX_DEPTH = 2,
        STATE_DIVS_NEAR = 8,
        STATE_DIVS_MID = 12,
        STATE_DIVS_FAR = 15,
        NUM_STATE = STATE_DIVS_NEAR + STATE_DIVS_MID + STATE_DIVS_FAR,
        DEST_DIR_DIVS = 72, // step: 5 degree
    };

    enum Flag {
        SAFETY = 0x0000,
        NEXT_TACKLABLE = 0x0001,
        NEXT_KICKABLE = 0x0002,
        TACKLABLE = 0x0004,
        KICKABLE = 0x0008,
        SELF_COLLISION = 0x0010,
        RELEASE_INTERFERE = 0x0020,
    };

    /*
      \struct State
      \brief class to represent a kick intermidiate state
    */
    struct State {
        int index_; //!< index of this point
        Vector2D pos_; //!< position relative to player's body
        double kick_rate_; //!< kick rate
        int flag_;

        State()
            : index_( -1 )
            , kick_rate_( 0.0 )
            , flag_( 0xFFFF )
          { }

        State( const int index,
               const Vector2D & pos,
               const double & kick_rate )
            : index_( index )
            , pos_( pos )
            , kick_rate_( kick_rate )
            , flag_( SAFETY )
          { }

    };

    /*!
      \struct Path
      \brief used as a heuristic knowledge. path representation between two states
     */
    struct Path {
        int origin_; //!< index of origin state
        int dest_; //!< index of destination state
        double max_speed_; //!< reachable ball max speed
        double power_; //!< kick power to generate max_speed_

        Path( const int origin,
              const int dest )
            : origin_( origin )
            , dest_( dest )
            , max_speed_( 0.0 )
            , power_( 1000.0 )
          { }

    };

    /*!
      \struct Sequence
      \brief simulated kick sequence
     */
    struct Sequence {
        int flag_; //!< safety level flags
        std::vector< Vector2D > pos_list_; //!< ball positions
        double speed_; //!< released ball speed
        double power_; //!< estimated last kick power
        double score_; //!< evaluated score of this sequence

        Sequence()
            : flag_( 0x0000 )
            , speed_( 0.0 )
            , power_( 10000.0 )
            , score_( 0.0 )
          { }
    };

    static
    double calc_near_dist( const PlayerType & player_type );
    static
    double calc_mid_dist( const PlayerType & player_type );
    static
    double calc_far_dist( const PlayerType & player_type );

    static
    Vector2D calc_max_velocity( const AngleDeg & target_angle,
                                const double & krate,
                                const Vector2D & ball_vel );

    static
    int get_nearest_index( const Vector2D & current_self_pos,
                           const AngleDeg & current_self_body,
                           const PlayerType & self_type,
                           const Vector2D & current_ball_pos );

private:

    //! static state list
    std::vector< State > M_state_list;

    //! sorted table
    std::vector< Path > M_tables[DEST_DIR_DIVS];

    //! current state cache
    State M_current_state;
    //! future state cache
    std::vector< State > M_state_cache[MAX_DEPTH];

    //! result kick sequences
    std::vector< Sequence > M_candidates;

    /*!
      \brief private constructor for singleton
     */
    KickSimulator();

    // not used
    KickSimulator( const KickSimulator & );
    const KickSimulator & operator=( const KickSimulator & );

private:
    void createTables();
    void createTable( const AngleDeg & angle,
                      std::vector< Path > & table );
    void createStateList();

    void updateState( const WorldModel & world );

    void createStateCache( const WorldModel & world );

    void checkCollisionAfterRelease( const WorldModel & world,
                                     const Vector2D & target_point,
                                     const double & first_speed );
    void checkInterfereAt( const WorldModel & world,
                           const int cycle,
                           State & state );
    void checkInterfereAfterRelease( const WorldModel & world,
                                     const Vector2D & target_point,
                                     const double & first_speed );
    void checkInterfereAfterRelease( const WorldModel & world,
                                     const Vector2D & target_point,
                                     const double & first_speed,
                                     const int cycle,
                                     State & state );

    bool simulateOneStep( const WorldModel & world,
                          const Vector2D & target_point,
                          const double & first_speed );
    bool simulateTwoStep( const WorldModel & world,
                          const Vector2D & target_point,
                          const double & first_speed );
    bool simulateThreeStep( const WorldModel & world,
                            const Vector2D & target_point,
                            const double & first_speed );

    void evaluate( const double & first_speed );

public:

    static
    KickSimulator & instance();

    bool simulate( const WorldModel & world,
                   const Vector2D & target_point,
                   const double & first_speed,
                   const int max_step,
                   Sequence & sequence );

    const
    std::vector< Sequence > & candidates() const
      {
          return M_candidates;
      }

};

}

#endif
